from dataclasses import dataclass
from typing import Optional


@dataclass
class MatchState:
    home_score: int
    away_score: int
    time: float
    current_period: int
    period_active: bool
    current_possession_team_id: Optional[str] = None

    @property
    def time_str(self):
        if not self.period_active:
            if self.current_period == 0:
                return "START"
            elif self.current_period == 1:
                return "RUST"
            elif self.current_period == 2:
                return "EINDE"
        else:
            minutes, seconds = divmod(int(self.time), 60)
            return f"{minutes:02d}:{seconds:02d}"

    def __repr__(self):
        return f"<MatchState time={self.time} score={self.home_score}-{self.away_score} active={self.period_active}>"



from .teamtv_object import TeamTVObject
from .resource_group.factory import factory as resource_group_factory


class Membership(TeamTVObject):
    @property
    def role_names(self):
        return self._role_names

    @property
    def resource_group(self):
        return self._resource_group

    @property
    def member_attributes(self):
        return self._member_attributes

    def _use_attributes(self, attributes: dict):
        self._role_names = attributes["roleNames"]
        self._member_attributes = attributes["memberAttributes"]

        self._resource_group = resource_group_factory(
            self._requester,
            dict(
                tenantId=attributes["tenantId"],
                resourceGroupId=attributes["resourceGroupId"],
                targetResourceId=attributes["targetResourceId"],
                targetResourceName=attributes["targetResourceName"],
            ),
        )

        super()._use_attributes(attributes)

    def __repr__(self):
        return f"<Membership resourceGroup={self.resource_group} role={self.role_names[0]}>"



from . import _ResourceGroup

from .capabilities import (
    _HasTeamsMixin,
    _HasSportingEventsMixin,
    _HasVideosMixin,
    _HasPersonsMixin,
)


class TeamResourceGroup(
    _ResourceGroup,
    _HasTeamsMixin,
    _HasSportingEventsMixin,
    _HasVideosMixin,
    _HasPersonsMixin,
):
    pass



from datetime import datetime
from typing import Literal, Optional

from pyteamtv.infra.requester import Requester
from ..list import List
from ..person import Person

from ..sporting_event import SportingEvent
from ..team import Team
from ..video import Video
from ..exercise import Exercise
from ..app_storage import AppStorageTokens


def iso8601(datetime_: datetime):
    assert datetime_.tzname() == "UTC", datetime_

    return datetime_.strftime("%Y-%m-%dT%H:%M:%S.%fZ")


class BaseMixin(object):
    _requester: Requester


class _HasStorage(BaseMixin):
    def get_app_storage_tokens(self):
        return AppStorageTokens(
            self._requester,
            self._requester.request("GET", "/apps/current/storageTokens"),
        )


class _HasTeamsMixin(BaseMixin):
    def get_teams(self):
        return List(Team, self._requester, "GET", "/teams")

    def create_team(self, name: str, sport_type: str, tags: dict = None):
        data = self._requester.request(
            "POST",
            "/teams",
            dict(name=name, sportType=sport_type, tags=tags if tags else {}),
        )
        return Team(self._requester, data)


class _HasExercisesMixin(BaseMixin):
    def get_exercises(self):
        return List(Exercise, self._requester, "GET", "/exercises")

    def create_exercise(
        self,
        title: str,
        exercise_template_id: str = "exercise-template-sv",
        tags: dict = None,
    ):
        data = self._requester.request(
            "POST",
            "/exercises",
            dict(
                title=title,
                exerciseTemplateId=exercise_template_id,
                tags=tags if tags else {},
            ),
        )
        return Exercise(self._requester, data)


class _HasSportingEventsMixin(BaseMixin):
    def get_sporting_events(self):
        return List[SportingEvent](
            SportingEvent, self._requester, "GET", "/sportingEvents"
        )

    def get_sporting_event(self, sporting_event_id: str):
        data = self._requester.request("GET", f"/sportingEvents/{sporting_event_id}")
        return SportingEvent(self._requester, data)

    def create_sporting_event(
        self,
        home_team: Team,
        away_team: Team,
        scheduled_at: datetime,
        tags: dict = None,
        type: str = "match",
    ):
        data = self._requester.request(
            "POST",
            "/sportingEvents",
            dict(
                type=type,
                name=f"{home_team.name} - {away_team.name}",
                homeTeamId=home_team.team_id,
                awayTeamId=away_team.team_id,
                scheduledAt=iso8601(scheduled_at),
                tags=tags if tags else {},
            ),
        )
        return SportingEvent(self._requester, data)


class _HasVideosMixin(BaseMixin):
    def get_videos(self):
        return List(Video, self._requester, "GET", "/videos")

    def get_video(self, video_id):
        data = self._requester.request("GET", f"/videos/{video_id}")

        return Video(self._requester, data)


class _HasPersonsMixin(BaseMixin):
    def get_persons(self):
        return List(Person, self._requester, "GET", "/persons")

    def create_person(
        self,
        first_name: str,
        last_name: str,
        tags: Optional[dict] = None,
        gender: Optional[Literal["male", "female"]] = None,
    ):
        data = self._requester.request(
            "POST",
            "/persons",
            {
                "firstName": first_name,
                "lastName": last_name,
                "tags": tags,
                "gender": gender,
            },
        )

        return Person(self._requester, data)


class _HasResourceGroupsMixin(BaseMixin):
    def get_resource_groups(self):
        from .factory import factory

        return List(factory, self._requester, "GET", "/resourceGroups")

    def join(self, resource_group_id: str):
        self._requester.request("POST", f"/resourceGroups/{resource_group_id}/join")


class _HasSharingGroupResourceGroupsMixin(BaseMixin):
    def get_sharing_group_resource_groups(self):
        return _HasResourceGroupsMixin.get_resource_groups(self)

    def create_sharing_group_resource_group(
        self,
        name: str,
        description: str,
        sport_type: str,
        visibility_type: str = "public",
        attributes: dict = None,
    ):
        from .factory import factory

        data = self._requester.request(
            "POST",
            "/resourceGroups",
            dict(
                name=name,
                sportType=sport_type,
                description=description,
                visibility_type=visibility_type,
                attributes=attributes if attributes else {},
            ),
        )
        return factory(self._requester, data)



from .person import Person
from .teamtv_object import TeamTVObject


class Player(TeamTVObject):
    @property
    def team(self):
        return self._team

    @property
    def number(self):
        return self._number

    @property
    def person(self):
        return self._person

    def _use_attributes(self, attributes: dict):
        self._person = Person(self._requester, attributes["person"])
        self._number = attributes["number"]
        self._team = attributes["team"]

        super()._use_attributes(attributes)

    def __repr__(self):
        return f"<Player name='{self.person.name}' team='{self.team.name}'>"



import math

from typing import List, Optional

from pyteamtv.models.observation_log import ObservationLog
from pyteamtv.models.person import Person
from pyteamtv.models.resource_group.team import TeamResourceGroup
from pyteamtv.models.sporting_event import SportingEvent
from pyteamtv.models.team import Team


def pol2cart(angle: float, distance: float):
    if angle > 90:
        angle -= 360
    angle += 90

    phi = (angle / 360) * 2 * math.pi
    rho = distance

    x = rho * math.cos(phi)
    y = rho * math.sin(phi)
    return x, y


class DataframeBuilder:
    def __init__(self, resource_group: TeamResourceGroup):
        self.teams = {team.team_id: team for team in resource_group.get_teams()}
        self.persons = {
            person.person_id: person for person in resource_group.get_persons()
        }
        self._requester = resource_group._requester

    def build_df(self, observation_logs: List[ObservationLog]):
        try:
            import pandas as pd
        except ImportError:
            raise Exception("You don't have pandas installed. Please install first")

        return pd.DataFrame.from_records(self.build_records(observation_logs))

    def _build_person_data(self, attributes: dict, key_prefix: Optional[str] = None):
        in_key = "person"
        out_prefix = ""
        if key_prefix:
            in_key = f"{key_prefix}Person"
            out_prefix = key_prefix + "_"

        pk_name = f"{in_key}Id"

        if in_key in attributes:
            person = {
                f"{out_prefix}person_id": attributes[pk_name],
                f"{out_prefix}first_name": attributes[in_key]["firstName"],
                f"{out_prefix}last_name": attributes[in_key]["lastName"],
                f"{out_prefix}number": attributes[in_key]["number"],
                f"{out_prefix}full_name": attributes[in_key]["firstName"]
                + " "
                + attributes[in_key]["lastName"],
            }
        elif pk_name in attributes:
            person_: Optional[Person] = self.persons.get(attributes[pk_name])
            if person_:
                person = {
                    f"{out_prefix}person_id": attributes[pk_name],
                    f"{out_prefix}first_name": person_.first_name,
                    f"{out_prefix}last_name": person_.last_name,
                    f"{out_prefix}full_name": person_.name,
                }
            else:
                person = dict()
        else:
            person = dict()
        return person

    def get_team_data(self, sporting_event: SportingEvent, attributes: dict):
        """
        1. Get it from the list of teams
        2. Get it from the attributes
        3. Get a generic one from the SportingEvent

        NOTE: this function has a side-effect: the self.teams dictionary will
              be updated when a team is not found and a fake team is created

        """

        team_id = attributes["teamId"]
        if sporting_event.home_team_id == team_id:
            ground = "home"
        else:
            ground = "away"

        if team_id in self.teams:
            team = self.teams[team_id]
        else:
            if "team" in attributes:
                # This can happen when a team is not shared, but the data is.
                # For certain leagues the data is entered from the club account
                # and send to the exchange. The teams entities are not shared
                team = Team(self._requester, attributes)
            else:
                home_team_name, away_team_name = sporting_event.name.split(" - ")
                team = Team(
                    self._requester,
                    {
                        "teamId": team_id,
                        "name": (
                            home_team_name if ground == "home" else away_team_name
                        ),
                    },
                )

            # It's a bad practice to update data in a get function
            # TODO: refactor
            self.teams[team_id] = team

        return dict(
            team_id=team.team_id,
            team_name=team.name,
            team_ground=ground,
            position=attributes.get("position"),
            team_name_full=team.full_name,
            team_key=team.key,
        )

    def build_records(self, observation_logs: List[ObservationLog]):
        observations = []
        skip_attributes = {
            "teamId",
            "personId",
            "team",
            "person",
            "inPerson",
            "inPersonId",
            "outPerson",
            "outPersonId",
            "opponentPersonId",
        }

        for observation_log in observation_logs:
            sporting_event = observation_log.sporting_event

            team = dict()
            for observation in observation_log:

                if observation.code in ("START-POSSESSION", "POSSESSION"):
                    team = self.get_team_data(sporting_event, observation.attributes)
                else:
                    person = self._build_person_data(observation.attributes)
                    opponent_person = self._build_person_data(
                        observation.attributes, "opponent"
                    )

                    attributes = dict()
                    attributes.update(team)
                    attributes.update(person)
                    attributes.update(opponent_person)
                    for k, v in observation.attributes.items():
                        if k not in skip_attributes:
                            attributes[k] = v

                    if "angle" in attributes and "distance" in attributes:
                        attributes["x"], attributes["y"] = pol2cart(
                            attributes["angle"], attributes["distance"]
                        )

                    observation_dict = dict(
                        sporting_event_id=sporting_event.sporting_event_id,
                        sporting_event_name=sporting_event.name,
                        sporting_event_scheduled_at=sporting_event.scheduled_at,
                        observation_id=observation.observation_id,
                        clock_id=observation.clock_id,
                        start_time=observation.start_time,
                        end_time=observation.end_time,
                        code=observation.code,
                        description=observation.description,
                        **attributes,
                    )

                    observations.append(observation_dict)

        return observations



from datetime import datetime, timezone
import time
from queue import Queue
from typing import Iterator, Tuple, Optional

from threading import Thread, Event as ThreadEvent

from .match.match_config import MatchConfig
from .match.match_state import MatchState
from .teamtv_object import TeamTVObject
from ..infra.match_state.event_store import QueueEventStore, Event
from ..infra.match_state.match_config_calculator import calculate_match_config
from ..infra.match_state.match_event_fetcher import match_event_fetcher
from ..infra.match_state.match_state_calculator import calculate_match_state


def utcnow() -> datetime:
    return datetime.fromtimestamp(time.time(), timezone.utc)


class EventStreamReader(Iterator):
    def get_last_event(self) -> Optional[Tuple[MatchConfig, MatchState, Event]]:
        events = self.event_store.get_events()

        timestamp = utcnow()  # TODO: or should this be occurredOn attribute of event

        while len(events) > self.cursor:
            state = calculate_match_state(events[: self.cursor + 1], timestamp)
            match_config = calculate_match_config(events[: self.cursor + 1], timestamp)
            item = events[self.cursor]
            self.cursor += 1
            if self.start_timestamp and item.occurred_on < self.start_timestamp:
                continue

            return match_config, state, item

        return None

    def __next__(self) -> Tuple[MatchConfig, MatchState, Event]:
        while True:
            event = self.get_last_event()
            if event:
                return event

            time.sleep(0.1)

    def __init__(self, poll_url: str, seek_to_start: bool = False):
        self.queue = Queue()
        self.start_timestamp = None if seek_to_start else utcnow()
        self.event_store = QueueEventStore(self.queue)
        self.should_stop = ThreadEvent()
        self.cursor = 0
        self.poll_url = poll_url

        self.thread = Thread(target=self._read_to_queue, daemon=True)
        self.thread.start()

    def _read_to_queue(self):
        match_event_fetcher(self.poll_url, self.queue, self.should_stop)

    def __iter__(self):
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.should_stop.set()
        self.thread.join()


class EventStream(TeamTVObject):
    @property
    def event_stream_id(self):
        return self._event_stream_id

    @property
    def endpoint_urls(self):
        return self._endpoint_urls

    def __repr__(self):
        return f"<EventStream event_stream_id={self.event_stream_id}>"

    def _use_attributes(self, attributes: dict):
        self._event_stream_id = attributes["eventStreamId"]
        self._endpoint_urls = attributes["endpointUrls"]

        super()._use_attributes(attributes)

    def open(self, seek_to_start: bool = False):
        return EventStreamReader(
            self.endpoint_urls["polling"], seek_to_start=seek_to_start
        )
